/*
 *  aiffreadfile.cpp
 *  A class for reading simple AIFF and AIFC files with uncompressed data.
 *
 *  Created by Alexander Thomas on 2008/01/22.
 *  Modified on 2014/01/11
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2014, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cstring>
#include <iostream>
using namespace std;

#include "pcmreadfile.h"
#include "aiffreadfile.h"

#define AIFF_IDSIZE    4
#define AIFF_SHORTSIZE 2
#define AIFF_LONGSIZE  4


AIFFReadFile::AIFFReadFile( const string &sFileName, bool bVerbose ) :
	PCMReadFile( sFileName, bVerbose )
{	
	m_bBigEndian = true;

	char id[AIFF_IDSIZE+1];
	m_istrAudioFile.read( id, AIFF_IDSIZE );
	id[AIFF_IDSIZE] = 0;
	if( strcmp(id,"FORM") != 0 ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Not a FORM (AIFF) file" << endl;
		return;
	}
	readValue( (char*)&(m_soundInfo.totalDataSize), AIFF_LONGSIZE );

	bool bAIFC=false;
	m_istrAudioFile.read( id, AIFF_IDSIZE );
	if( strcmp(id,"AIFC") == 0 )
		bAIFC = true;
	else if( strcmp(id,"AIFF") != 0 ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Not a FORM (AIFF) file" << endl;
		return;
	}

	if( ! scanFile( m_istrAudioFile, "COMM", AIFF_IDSIZE ) ) {
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Not a valid AIFF file (COMM chunk not found)" << endl;
		return;
	}
	// skip indicator of comm chunk length (should be 0x12 for AIFF, variable for AIFC)
	m_istrAudioFile.ignore(AIFF_LONGSIZE);
	readValue( (char*)&(m_soundInfo.numChannels), AIFF_SHORTSIZE );
	readValue( (char*)&(m_soundInfo.numSampleFrames), AIFF_LONGSIZE );
	readValue( (char*)&(m_soundInfo.sampleSize), AIFF_SHORTSIZE );
	if( m_soundInfo.sampleSize != 16) {
		// TODO: allow other sample sizes!
		m_fileState = Bad;
		if(m_bVerbose) cerr << "Error: Only 16-bit AIFF files supported (this file is " << m_soundInfo.sampleSize << "-bit)" << endl;
		return;
	}
	extended_t temp;
	m_istrAudioFile.read( (char*)temp, sizeof(extended_t) );
	m_soundInfo.sampleRate = IEEE754ToDouble(temp);
	if(bAIFC) {
		m_istrAudioFile.read( id, AIFF_IDSIZE );
		if( strcmp(id,"sowt") == 0 ) // 'codec' for little endian 16-bit
			m_bBigEndian = false;
		else if( strcmp(id,"NONE") != 0 ) {
			m_fileState = Bad;
			if(m_bVerbose)
				cerr << "Error: AIFC compression type `" << id << "' not supported" << endl;
			return;
		}
	}
	
	scanFile( m_istrAudioFile, "SSND", AIFF_IDSIZE );
	// skip indicator of chunk length, offset & blockSize
	m_istrAudioFile.ignore(3*AIFF_LONGSIZE);
	
	m_audioStart = m_istrAudioFile.tellg();
	m_fileState  = Good;
}
