/*
 *  endianconv.h
 *  Tools for endianness conversion
 *
 *  This is not optimized at all, it just does the job.
 *
 *  Created by Alexander Thomas on 2008/01/24.
 *  Last edited: 2014/01/11
 *
 */

#ifndef ENDIANCONV_H
#define ENDIANCONV_H

/* TODO: use a more universal way. As with endianness, there is an annoying lack
 * of cross-platform methods to probe platform capabilities. */
#define UInt16 unsigned short
#define UInt32 unsigned long

/* There is no foolproof cross-platform way to determine endianness.
 * The following hack first tests if OSX-style defines are available,
 *   if not, it tries to include the glibc endian.h file. */
#if !defined(__LITTLE_ENDIAN__) && !defined(__BIG_ENDIAN__)
	#include <endian.h>
	#ifdef __BYTE_ORDER
		#if __BYTE_ORDER == __BIG_ENDIAN
			#define __BIG_ENDIAN__ 1
			#define __LITTLE_ENDIAN__ 0
		#elif __BYTE_ORDER == __LITTLE_ENDIAN
			#define __BIG_ENDIAN__ 0
			#define __LITTLE_ENDIAN__ 1
		#elif __BYTE_ORDER == __PDP_ENDIAN
			#warning WARNING: PDP_ENDIAN not supported!
		#endif
	#endif

	#if !defined(__LITTLE_ENDIAN__) && !defined(__BIG_ENDIAN__)
		#warning WARNING: cannot determine endianness.
	#endif
#endif


/* General endianness swap: reverse byte order in a block of wLength bytes. */
inline void swapEndian( void* pBuff, unsigned int wLength )
{
	char* pcBuff = (char*)pBuff;
	for( unsigned int j=0; j<(wLength>>1); j++ ) {
		char c = pcBuff[j];
		pcBuff[j] = pcBuff[wLength-j-1];
		pcBuff[wLength-j-1] = c;
	}
}

/* Treat buffer as a Big endian number, i.e.
 * reverse byte order if this is a little endian machine. */
inline void getFromBigEndian( void* pBuff, unsigned int wLength )
{
#if __LITTLE_ENDIAN__
	swapEndian( pBuff, wLength);
#endif
}

/* Treat buffer as a little endian number, i.e.
 * reverse byte order if this is a Big endian machine. */
inline void getFromLittleEndian( void* pBuff, unsigned int wLength )
{
#if __BIG_ENDIAN__
	swapEndian( pBuff, wLength);
#endif
}


/* More efficient fixed-size versions of the generic above methods: */
/* If this is a little Endian machine, swap bytes of the argument. */
inline void getFromBigEndian16Bit( void* p )
{
#if __LITTLE_ENDIAN__
	char *buf = (char*)p;
	char c = buf[0];
	buf[0] = buf[1];
	buf[1] = c;
#endif
}

/* If this is a Big Endian machine, swap bytes of the argument. */
inline void getFromLittleEndian16Bit( void* p )
{
#if __BIG_ENDIAN__
	char *buf = (char*)p;
	char c = buf[0];
	buf[0] = buf[1];
	buf[1] = c;
#endif
}


/* Swap nPairs byte pairs in buffer if this is a little endian machine,
 * otherwise do nothing. */
inline void get16BitArrayFromBigEndian( void* buff, unsigned long nPairs )
{
#if __LITTLE_ENDIAN__
	char* t = (char*)buff;
	for( unsigned long i=0; i<nPairs; i++ ) {
		char c = t[i<<1];
		t[i<<1] = t[(i<<1)+1];
		t[(i<<1)+1] = c;
	}
#endif
}


/* Swap nPairs byte pairs in buffer if this is a Big endian machine,
 * otherwise do nothing. */
inline void get16BitArrayFromLittleEndian( void* buff, unsigned long nPairs )
{
#if __BIG_ENDIAN__
	char* t = (char*)buff;
	for( unsigned long i=0; i<nPairs; i++ ) {
		char c = t[i<<1];
		t[i<<1] = t[(i<<1)+1];
		t[(i<<1)+1] = c;
	}
#endif
}	


#endif // ENDIANCONV_H
