/*
 * SoundStats
 * A program for gathering statistics about mono or stereo sound files.
 * The program will construct a histogram for the occurrence of each sample
 *   value in all input files, as well as histograms of how often a piece
 *   of clipping waveform of a certain length was encountered. Results can
 *   be printed as console output or written to tab-delimited text files.
 *
 * Created by Alexander Thomas on 2011/01/04
 * Last modification: 2014/01/11
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2014, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* TODO:
 * - Try to extract a single reasonable 'dynamic range' figure (from the sample histo and/or replayGain)
 * - Add power spectrum calculation
 * - Add AAC decoding
 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <string>
using namespace std;
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <cassert>
#include <ctype.h>
#include <sys/stat.h>

#include "aiffreadfile.h"
#include "wavreadfile.h"
// Must include this for in case !(MP3_SUPPORT || VORBIS_SUPPORT)
#include "decoders.h"
#ifdef MP3_SUPPORT
#include "mp3decoder.h"
#endif
#ifdef VORBIS_SUPPORT
#include "vorbisdecoder.h"
#endif
#include "soundstats.h"
// This is actually C, but the .h file is C++ savvy
#include "gain_analysis.h"

#define SOUNDSTATS_VERSION "1.1.2"
#define DEFAULT_CLIPTHRESH 0.99f
// Defaults for console output
#define DEFAULT_HEIGHT 21
#define DEFAULT_WIDTH  76

//#define DEBUG

#define SAMPLES_PER_CHUNK 32768

static bool m_bVerbose=true;


void printUsage( const char *pName, bool bExtended=false )
{
	if(bExtended) {
		cout << "SoundStats " << SOUNDSTATS_VERSION << " by Alexander Thomas" << endl;
		cout << "Collects statistics about sound files." << endl;
	}
	cout << "Usage:\n"
	     << "  " << pName << " [options] file[s]" << endl;
	cout << "Options:\n"
	     << "  -h: show usage info and help, and exit.\n"
	     << "  -V: disable verbose mode.\n"
	     << "  -s: output separate statistics for left and right if input is stereo.\n"
	     << "  -R: disable ReplayGain computation.\n"
	     << "  -o O: write statistics to output files with names starting in O.\n"
	     << "  -t T: consider a sample to be clipping if it is at or above a fraction T of\n"
	     << "        the maximum level  (default: " << DEFAULT_CLIPTHRESH << ")\n"
	     << "  -l L: set number of lines in console output histogram to L (default: " << DEFAULT_HEIGHT <<").\n"
	     << "        The clipping histogram is L/2 high. Set to 0 to disable ASCII histogram.\n"
	     << "  -w W: set width of console output histogram to W (default: " << DEFAULT_WIDTH << ")." << endl;
	if(bExtended) {
		cout << "The output files, if -o is used, will consist of a `-samples.txt' file that\n"
		     << "  contains for each quantized sample value the number of times it occurred; and\n"
		     << "  a `-clipping.txt' file that contains values x,y such that y is the number of\n"
		     << "  times a set of x consecutive samples were encountered that were considered to\n"
		     << "  be clipping. If no such samples were encountered, this file will not be\n"
		     << "  written." << endl;
		cout << "Mind that for lossy input file formats it is possible that sample values beyond\n"
		     << "  the 16-bit range will be reported, or in other words below or above [-1,1] in\n"
		     << "  a normalized range. This is due to the way lossy compression works. For\n"
		     << "  properly mastered music with sufficient headroom this should occur rarely but\n"
		     << "  if the source material was already hugging the extremal values then the\n"
		     << "  compressed waveform will certainly overshoot." << endl;
		cout << "Compiled with support for: WAV, AIFF";
#ifdef VORBIS_SUPPORT
		cout << ", OGG Vorbis";
#endif
#ifdef MP3_SUPPORT
		cout << ", MP3";
#endif
		cout << "." << endl;
	}
}

void error_exit( const char *msg, int val, const char *pName )
{
	cerr << msg << endl;
	if(pName)
		printUsage(pName);
	exit(val);
}


/*--- File routines ---*/
// Return the size of a file.
unsigned long getFileSize( const char *fName )
{
	struct stat buffer;
	if( stat(fName, &buffer) != 0 ) {
		cerr << "Error while opening file `" << fName << "'!" << endl;
		exit(1);
	}
	return buffer.st_size;
}		

// Return the file extension, including the '.'
string getExtension( const char *fileName )
{
	string sExt(fileName);
	size_t nPos = sExt.rfind('.');
	if( nPos == string::npos )
		return string("");
	sExt = sExt.substr(nPos);
	for( unsigned int i=0; i<sExt.length(); i++ )
		sExt[i] = tolower(sExt[i]);
	return sExt;
}

// Write sample histogram file
void writeHistoFile( const string &sFileName, const vector<unsigned long> &vHistoNeg,
                     const vector<unsigned long> &vHistoPos )
{
	ofstream oFile( sFileName.c_str() );
	for( long i=vHistoNeg.size()-1; i>=0; i-- )
		oFile << -i-1 << '\t' << vHistoNeg[i] << "\n";
	for( unsigned int i=0; i<vHistoPos.size(); i++ )
		oFile << i << '\t' << vHistoPos[i] << "\n";
	oFile.close();
}

/*--- Other tools ---*/
// Return milliseconds for given number of samples at sample rate, rounded to 2 digits
float samp2mSec( int nSamples, long nSampleRate )
{
	return .01f*roundf(100000.0*nSamples/nSampleRate);
}

// Return the maximum value in the vector.
template <class dType>
dType vectorMax( const vector<dType> &vNumbers )
{
	assert( vNumbers.size() > 0 );
	dType nMax=vNumbers[0];
	for( unsigned int i=1; i<vNumbers.size(); i++ ) {
		if( vNumbers[i] > nMax )
			nMax = vNumbers[i];
	}
	return nMax;
}

// Returns the total power in the histograms, as well as the sample value in
// nPeakPos for which the most samples occur.
double getPowerAndPeak( int &nPeakPos,
                        const vector<unsigned long> &vRawHistoPos,
                        const vector<unsigned long> &vRawHistoNeg )
{
	double dTotPower=0.0;
	unsigned long nPeakValue=vRawHistoPos[0];
	nPeakPos=0;

	for( unsigned int i=1; i<vRawHistoPos.size(); i++ ) {
		double dSampVal = (double)i/32768;
		dTotPower += dSampVal*dSampVal*vRawHistoPos[i];
		if( vRawHistoPos[i] > nPeakValue ) {
			nPeakValue = vRawHistoPos[i];
			nPeakPos = i;
		}
	}
	for( unsigned int i=0; i<vRawHistoNeg.size(); i++ ) {
		double dSampVal = (double)(i+1)/32768;
		dTotPower += dSampVal*dSampVal*vRawHistoNeg[i];
		if( vRawHistoNeg[i] > nPeakValue ) {
			nPeakValue = vRawHistoNeg[i];
			nPeakPos = -(i+1);
		}
	}

	return dTotPower;
}


// Update necessary variables and return true if sample rate has changed.
// nGlobalSampleRate is the last seen value.
// bReplayGain will be set to false if the sample rate is not supported.
bool checkSampleRateChange( long nSampleRate, long &nGlobalSampleRate, bool &bReplayGain )
{
	if( nSampleRate != nGlobalSampleRate ) {
		int nRGResult = INIT_GAIN_ANALYSIS_OK;
		if( nGlobalSampleRate == 0 ) {
			if(bReplayGain)
				nRGResult = InitGainAnalysis(nSampleRate);
		}
		else {
			cerr << "Warning: different sample rates, assuming last sample rate. Total time values may be incorrect." << endl;
			if(bReplayGain)
				nRGResult = ResetSampleFrequency(nSampleRate);
		}

		nGlobalSampleRate = nSampleRate;

		if( nRGResult == INIT_GAIN_ANALYSIS_ERROR ) {
			bReplayGain = false;
			cerr << "Error: disabling ReplayGain due to unsupported sampling rate." << endl;
		}

		return true;
	}

	return false;
}


// Scale all values in the array(s)
void scaleSamples( float *pSamplesLeft, float *pSamplesRight, unsigned long nSamples, float fScale )
{
	if( pSamplesRight != NULL ) {
		for( unsigned long s = 0; s < nSamples; s++ ) {
			pSamplesLeft[s]  *= fScale;
			pSamplesRight[s] *= fScale;
		}
	}
	else {
		for( unsigned long s = 0; s < nSamples; s++ )
			pSamplesLeft[s] *= fScale;
	}
}


// Remove trailing spaces from every row in the given ASCII art image.
void squashASCIIImage( vector< vector<char> > &vvcImage )
{
	for( unsigned int i=0; i<vvcImage.size(); i++ ) {
		unsigned int nLen = vvcImage[i].size();
		while( nLen > 0 ) {
			if( vvcImage[i][nLen-1] != ' ' )
				break;
			nLen--;
		}
		if( vvcImage[i].size() != nLen )
			vvcImage[i].resize(nLen);
	}
}

// Render an upright ASCII image of the given histogram.
// The resulting image will be as wide (length of inner vectors) as the length of vnHisto,
//   and nHeight pixels tall (length of outer vector).
// The bins will be scaled by nHeight/nScale. Any element larger than zero will be printed
//   as at least one, regardless of how small it is.
// If bFlip, the image will be flipped horizontally.
vector< vector<char> > renderHisto( vector<unsigned long> vnHisto,
                                    unsigned long nScale,
                                    unsigned long nHeight,
                                    bool bFlip=false )
{
	// Start out by creating an empty 'image'
	unsigned int nWidth = vnHisto.size();
	vector<char> vEmpty( nWidth, ' ' );
	vector< vector<char> > vvcOut( nHeight, vEmpty );

	// Draw histogram with '*'
	for( unsigned int i=0; i<nWidth; i++ ) {
		unsigned int nX = i;
		if(bFlip)
			nX = nWidth-1-i;
		unsigned int nVal = (unsigned int)round( nHeight*(float)vnHisto[i]/nScale );
		if( nVal == 0 && vnHisto[i] > 0 )
			nVal = 1;
		for( unsigned int j=0; j<nVal; j++ )
			vvcOut[nHeight-1-j][nX] = '*';
	}

	return vvcOut;
}


// Round to two decimals
double round2Dec( double x )
{
	x *= 100.0;
	return 0.01*round(x);
}


/*--- Main program ---*/
int main( int argc, char *argv[] )
{
	unsigned int nConsoleWidth=DEFAULT_WIDTH, nConsoleHeight=DEFAULT_HEIGHT;
	bool bMonoOutput=true, bVerbose=true, bReplayGain=true;
	char *sOutFile=NULL;
	vector<char*> vFiles;
	float fClipThresh=0.99f;

	// Parse arguments
	int nArg = 0;
	while( ++nArg < argc ) {
		if( argv[nArg][0] != '-' )
			vFiles.push_back(argv[nArg]);
		else {
			char *arg = argv[nArg];
			int nLen = strlen(arg);
			for( int j=1; j<nLen; j++ ) {
				switch( arg[j] ) {
				case 'h':
					printUsage( argv[0], true );
					exit(0);
					break;
				case 'o':
					if( nArg+1 == argc )
						error_exit( "Error: missing argument after -o.", 1, argv[0] );
					sOutFile = argv[++nArg];
					break;
				case 'R':
					bReplayGain = false;
					break;
				case 's':
					bMonoOutput = false;
					break;
				case 't':
					if( nArg+1 == argc || sscanf(argv[++nArg], "%g", &fClipThresh) != 1 )
						error_exit( "Error: -t expects a floating-point number als next argument.", 1, argv[0] );
					break;
				case 'l':
					if( nArg+1 == argc || sscanf(argv[++nArg], "%u", &nConsoleHeight) != 1 )
						error_exit( "Error: -l expects an integer als next argument.", 1, argv[0] );
					break;
				case 'w':
					if( nArg+1 == argc || sscanf(argv[++nArg], "%u", &nConsoleWidth) != 1 )
						error_exit( "Error: -w expects an integer als next argument.", 1, argv[0] );
					break;
				case 'V':
					bVerbose = false;
					break;
				default:
					cerr << "Warning: unknown option -" << arg[j] << endl;
					break;
				}
			}
		}
	}

	unsigned int nFiles = vFiles.size();
	if( nFiles == 0 )
		error_exit( "Not enough arguments", 1, argv[0] );


	SoundStats statsLeft, statsRight;
	int  nGlobalChannels=0;
	long nGlobalSampleRate=0;
	unsigned long nTotSamples=0;
	statsLeft.setClipThreshold(fClipThresh);
	statsRight.setClipThreshold(fClipThresh);

	for( unsigned int f=0; f<vFiles.size(); f++ ) {
		if(bVerbose)
			cout << "Analyzing `" << vFiles[f] << "'..." << endl;
		struct outBuffer samples = {false, NULL, NULL, 0, 0, 0};

		// I could make one 'Decoder' superclass and derive all other decoders from it.
		// TODO: allow loading other formats, like AAC
		string sExt = getExtension(vFiles[f]);
		if( sExt == ".aif" || sExt == ".aiff" || sExt == ".wav" ) {
			PCMReadFile *pDecoder=NULL;
			if( sExt == ".aif" || sExt == ".aiff" )
				pDecoder = new AIFFReadFile( string(vFiles[f]), m_bVerbose );
			else
				pDecoder = new WAVReadFile( string(vFiles[f]), m_bVerbose );
			
			if( pDecoder->getFileState() != Good ) {
				cerr << "WARNING: Could not load PCM file `" << vFiles[f] << "', skipping" << endl;
				continue;
			}
			int  nChannels = pDecoder->getNumChannels();
			long nSampleRate = (long)round( pDecoder->getSampleRate() );
			if(bVerbose)
				cout << "  Channels: " << nChannels << ", rate: " << nSampleRate << endl;

			if( nGlobalChannels == 0 )
				nGlobalChannels = nChannels;
			else if( nChannels > nGlobalChannels )
				cerr << "Warning: more channels than previous file(s), ignoring" << endl;
			else if( nChannels < nGlobalChannels ) {
				cerr << "Warning: fewer channels than previous file(s), dropping extra channel statistics" << endl;
				nGlobalChannels = nChannels;
			}
			checkSampleRateChange( nSampleRate, nGlobalSampleRate, bReplayGain );

			short **ppnSamples = new short*[nChannels];
			for( int c=0; c<nChannels; c++ )
				ppnSamples[c] = new short[SAMPLES_PER_CHUNK];
			
			// Do the analysis
			unsigned long nSamplesRead=0;
			while( (nSamplesRead = pDecoder->readSampleFrames( ppnSamples, SAMPLES_PER_CHUNK )) ) {
				nTotSamples += nSamplesRead;
				statsLeft.collectStats( ppnSamples[0], nSamplesRead );
				if( nGlobalChannels > 1 )
					statsRight.collectStats( ppnSamples[1], nSamplesRead );

				if(bReplayGain) {
					// Need to convert the samples to floats for this
					samples.pSamplesLeft = new float[nSamplesRead];
					if( nGlobalChannels > 1 ) {
						samples.pSamplesRight = new float[nSamplesRead];
						for( unsigned int s=0; s<nSamplesRead; s++ ) {
							// No need to scale, AnalyzeSamples expects a range of [-32767,32767]
							samples.pSamplesLeft[s]  = (float)ppnSamples[0][s];
							samples.pSamplesRight[s] = (float)ppnSamples[1][s];
						}
						AnalyzeSamples( samples.pSamplesLeft, samples.pSamplesRight, nSamplesRead, 2 );
						delete [] samples.pSamplesRight;
					}
					else {
						for( unsigned int s=0; s<nSamplesRead; s++ )
							samples.pSamplesLeft[s]  = (float)ppnSamples[0][s];
						AnalyzeSamples( samples.pSamplesLeft, NULL, nSamplesRead, 1 );
					}
					delete [] samples.pSamplesLeft;
				}
			}

			for( int c=0; c<nChannels; c++ )
				delete [] ppnSamples[c];
			delete [] ppnSamples;
		}
#ifdef VORBIS_SUPPORT
		else if( sExt == ".ogg" ) {
			VorbisDecoder decoder;
			if( ! decoder.init(vFiles[f]) ) {
				cerr << "WARNING: Could not decode file `" << vFiles[f] << "', skipping" << endl;
				continue;
			}
			int  nChannels = decoder.getNumChannels();
			long nSampleRate = decoder.getSampleRate();
			if(bVerbose)
				cout << "  Channels: " << nChannels << ", rate: " << nSampleRate << endl;

			if( nGlobalChannels == 0 )
				nGlobalChannels = nChannels;
			else if( nChannels > nGlobalChannels )
				cerr << "Warning: more channels than previous file(s), ignoring" << endl;
			else if( nChannels < nGlobalChannels ) {
				cerr << "Warning: fewer channels than previous file(s), dropping extra channel statistics" << endl;
				nGlobalChannels = nChannels;
			}
			checkSampleRateChange( nSampleRate, nGlobalSampleRate, bReplayGain );

			samples.nChannels = nChannels;
			samples.pSamplesLeft = new float[SAMPLES_PER_CHUNK];
			if( nChannels > 1 )
				samples.pSamplesRight = new float[SAMPLES_PER_CHUNK];
		
			// Do the analysis
			while( decoder.getSamples( samples, SAMPLES_PER_CHUNK ) ) {
				nTotSamples += samples.nSamples;
				statsLeft.collectStats( samples.pSamplesLeft, samples.nSamples );
				if( nGlobalChannels > 1 ) {
					statsRight.collectStats( samples.pSamplesRight, samples.nSamples );
					if(bReplayGain) {
						scaleSamples( samples.pSamplesLeft, samples.pSamplesRight, samples.nSamples, 32767.0f );
						AnalyzeSamples( samples.pSamplesLeft, samples.pSamplesRight, samples.nSamples, 2 );
					}
				}
				else {
					if(bReplayGain) {
						scaleSamples( samples.pSamplesLeft, NULL, samples.nSamples, 32767.0f );
						AnalyzeSamples( samples.pSamplesLeft, NULL, samples.nSamples, 1 );
					}
				}
			}
#ifdef DEBUG
			decoder.printStatus();
#endif

			delete [] samples.pSamplesLeft;
			delete [] samples.pSamplesRight;
		}
#endif // VORBIS_SUPPORT
#ifdef MP3_SUPPORT
		else if( sExt == ".mp3" ) {
			// TODO: could include ID3 junk in MP3Decoder, so I don't need to bother
			// about it and can make VorbisDecoder and MP3Decoder subclasses of
			// superclass 'Decoder'.
			long nId3Length = 0;
			bool bHasV1=false, bHasV2=false;

			if( ! MP3Decoder::getID3Info( vFiles[f], nId3Length, bHasV1, bHasV2 ) )
				error_exit( "File could not be read or is empty.", 1, NULL );
			if(m_bVerbose) {
				if(bHasV1) cout << "  File has an ID3v1 tag." << endl;
				if(bHasV2) cout << "  File has an ID3v2 tag." << endl;
			}
				
			unsigned long fileLength = getFileSize(vFiles[f]);
			unsigned long decodeLength = fileLength;
			FILE *theFile = fopen( vFiles[f], "rb" );
			if( theFile == NULL )
				error_exit( "Could not open input file", 1, NULL );
		
			if( nId3Length > 0 ) {
				fseek( theFile, nId3Length, SEEK_SET );
				decodeLength -= nId3Length;
			}
			if(bHasV1)
				decodeLength -= 128;

			MP3Decoder decoder( theFile, decodeLength );
			int  nChannels = decoder.getNumChannels();
			long nSampleRate = decoder.getSampleRate();
			if(bVerbose)
				cout << "  Channels: " << nChannels << ", rate: " << nSampleRate << endl;

			if( nGlobalChannels == 0 )
				nGlobalChannels = nChannels;
			else if( nChannels > nGlobalChannels )
				cerr << "Warning: more channels than previous file(s), ignoring" << endl;
			else if( nChannels < nGlobalChannels ) {
				cerr << "Warning: fewer channels than previous file(s), dropping extra channel statistics" << endl;
				nGlobalChannels = nChannels;
			}
			checkSampleRateChange( nSampleRate, nGlobalSampleRate, bReplayGain );

			samples.nChannels = nChannels;
			samples.pSamplesLeft = new float[SAMPLES_PER_CHUNK];
			if( nChannels > 1 )
				samples.pSamplesRight = new float[SAMPLES_PER_CHUNK];
		
			// Do the analysis
			while( decoder.getSamples( samples, SAMPLES_PER_CHUNK ) ) {
				nTotSamples += samples.nSamples;
				statsLeft.collectStats( samples.pSamplesLeft, samples.nSamples );
				if( nGlobalChannels > 1 ) {
					statsRight.collectStats( samples.pSamplesRight, samples.nSamples );
					if(bReplayGain) {
						scaleSamples( samples.pSamplesLeft, samples.pSamplesRight, samples.nSamples, 32767.0f );
						AnalyzeSamples( samples.pSamplesLeft, samples.pSamplesRight, samples.nSamples, 2 );
					}
				}
				else {
					if(bReplayGain) {
						scaleSamples( samples.pSamplesLeft, NULL, samples.nSamples, 32767.0f );
						AnalyzeSamples( samples.pSamplesLeft, NULL, samples.nSamples, 1 );
					}
				}
			}
			fclose(theFile);
#ifdef DEBUG
			decoder.printStatus();
#endif

			delete [] samples.pSamplesLeft;
			delete [] samples.pSamplesRight;
		}
#endif // MP3_SUPPORT
		else {
			cerr << "Error: no support for file of type `" << sExt << "'" << endl;
			continue;
		}

		if(bReplayGain) {
			// Apparently this must be called at the end of analyzing a file to finalize stuff
			// even if we don't want to know the track gain value
			float fTitleGain = GetTitleGain();
			if( bVerbose && vFiles.size() > 1 ) {
				string sPlus( fTitleGain > 0 ? "+" : "" );
				cout << "  ReplayGain: " << sPlus << fTitleGain << " dB" << endl;
			}
		}
	} // end file loop

	if( nGlobalChannels > 2 ) {
		cerr << "No support for more than 2 channels, only the first two channels were analyzed" << endl;
		nGlobalChannels = 2;
	}
	float dur = (float)nTotSamples/nGlobalSampleRate;
	cout << "Decoded a total of " << nTotSamples << " samples or " << dur << " seconds." << endl;

	statsLeft.finish();
	statsRight.finish();

	if( nTotSamples == 0 || nGlobalChannels == 0 ) {
		cout << "No data, exiting" << endl;
		return 0;
	}

	if( bVerbose && !bMonoOutput && nGlobalChannels == 1 )
		cout << "Note: stereo mode requested for mono input, output will be mono." << endl;

	bool bStereo = (nGlobalChannels > 1 && !bMonoOutput);

	if(!bStereo) {
		float fMin = statsLeft.getMinValue(),
		      fMax = statsLeft.getMaxValue();
		if( nGlobalChannels > 1 ) {
			if( statsRight.getMinValue() < fMin )
				fMin = statsRight.getMinValue();
			if( statsRight.getMaxValue() > fMax )
				fMax = statsRight.getMaxValue();
		}
		cout << "Min/max value: " << fMin << " " << fMax << endl;
	}
	else {
		cout << "Min/max value L: " << statsLeft.getMinValue() << " "
		     << statsLeft.getMaxValue() << endl;
		cout << "Min/max value R: " << statsRight.getMinValue() << " "
		     << statsRight.getMaxValue() << endl;
	}

	// Get histograms of clipping samples
	vector<unsigned long> vClipHistoL = statsLeft.getClipHistogram();
	vector<unsigned long> vClipHistoR = statsRight.getClipHistogram();;
	if(!bStereo) {
		if( nGlobalChannels > 1 ) {
			if( vClipHistoR.size() > vClipHistoL.size() )
				vClipHistoL.resize( vClipHistoR.size(), 0 );
			for( unsigned int i=0; i<vClipHistoR.size(); i++ )
				vClipHistoL[i] += vClipHistoR[i];
		}
		cout << "Maximum clip length: " << vClipHistoL.size() << " ("
		     << samp2mSec( vClipHistoL.size(), nGlobalSampleRate ) << " mSec)" << endl;
	}
	else {
		cout << "Maximum clip length L: " << vClipHistoL.size() << " ("
		     << samp2mSec( vClipHistoL.size(), nGlobalSampleRate ) << " mSec)" << endl;
		cout << "Maximum clip length R: " << vClipHistoR.size() << " ("
		     << samp2mSec( vClipHistoR.size(), nGlobalSampleRate ) << " mSec)" << endl;
	}
	unsigned long long nNumClips = 0;
	for( unsigned long i=0; i<vClipHistoL.size(); i++ )
		nNumClips += i*vClipHistoL[i];
	if(bStereo) {
		for( unsigned long i=0; i<vClipHistoR.size(); i++ )
			nNumClips += i*vClipHistoR[i];
	}
	double dClipRatio = (double)nNumClips/(nTotSamples*nGlobalChannels);
	cout << dClipRatio*100.0 << "% of all samples clip" << endl;

	// Get histograms for samples above and below the 'zero' (silence) line.
	vector<unsigned long> vRawHistoPosL, vRawHistoNegL;
	vector<unsigned long> vRawHistoPosR, vRawHistoNegR;
	statsLeft.getHistogram( vRawHistoPosL, vRawHistoNegL );
	if( nGlobalChannels > 1 ) {
		statsRight.getHistogram( vRawHistoPosR, vRawHistoNegR );

		if(bMonoOutput) {
			if( vRawHistoPosR.size() > vRawHistoPosL.size() )
				vRawHistoPosL.resize( vRawHistoPosR.size(), 0 );
			if( vRawHistoNegR.size() > vRawHistoNegL.size() )
				vRawHistoNegL.resize( vRawHistoNegR.size(), 0 );
			for( unsigned int i=0; i<vRawHistoPosR.size(); i++ )
				vRawHistoPosL[i] += vRawHistoPosR[i];
			for( unsigned int i=0; i<vRawHistoNegR.size(); i++ )
				vRawHistoNegL[i] += vRawHistoNegR[i];
		}
	}

	// Determine coverage
	unsigned long nNonZero=0;
	if(bStereo) {
		for( unsigned int i=0; i<32768; i++ ) {
			if( vRawHistoPosL[i] || vRawHistoPosR[i] ) nNonZero++;
			if( vRawHistoNegL[i] || vRawHistoNegR[i] ) nNonZero++;
		}
	}
	else {
		for( unsigned int i=0; i<32768; i++ ) {
			if( vRawHistoPosL[i] ) nNonZero++;
			if( vRawHistoNegL[i] ) nNonZero++;
		}
	}
	cout.precision(2);
	cout << "The file" << (nFiles > 1 ? "s" : "") << " use" << (nFiles > 1 ? " " : "s ")
	     << fixed << nNonZero*100.0/65536 << "% of all 16-bit values, an effective "
	     << log2(nNonZero) << " bits." << endl;

	// Find peak and calculate power
	int nPeakPosL=0;
	double dAvgPowerL = getPowerAndPeak( nPeakPosL, vRawHistoPosL, vRawHistoNegL );
	float fPeakPosL = (float)nPeakPosL/32768;
	cout.precision(5);
	if(!bStereo) {
		dAvgPowerL /= nTotSamples*nGlobalChannels;
		double dAvgDBmL = 10.0*log10(dAvgPowerL*1000.0);
		cout << "Most occurring value: " << nPeakPosL << " (" << fPeakPosL << ")" << endl;
		cout << "Average power: " << dAvgPowerL << " (" << setprecision(2) << dAvgDBmL << setprecision(5) << " dBm)" << endl;
	}
	else {
		int nPeakPosR=0;
		double dAvgPowerR = getPowerAndPeak( nPeakPosR, vRawHistoPosR, vRawHistoNegR );
		dAvgPowerL /= nTotSamples;
		dAvgPowerR /= nTotSamples;
		double dAvgDBmL = 10.0*log10(dAvgPowerL*1000.0),
		       dAvgDBmR = 10.0*log10(dAvgPowerR*1000.0);
		float fPeakPosR = (float)nPeakPosR/32768;
		cout << "Most occurring value L: " << nPeakPosL << " (" << fPeakPosL << "); R: "
		     << nPeakPosR << " (" << fPeakPosR << ")" << endl;
		cout << "Average power L: " << dAvgPowerL << " (" << setprecision(2) << dAvgDBmL << setprecision(5) << " dBm); R: "
		     << dAvgPowerR << " (" << setprecision(2) << dAvgDBmR << setprecision(5) << " dBm)" << endl;
	}

	if(bReplayGain) {
		float fAlbumGain = GetAlbumGain();
		cout.unsetf(ios::floatfield);
		string sPlus( fAlbumGain > 0 ? "+" : "" );
		cout << "ReplayGain: " << sPlus << fAlbumGain << " dB" << endl;
	}
	
	/*===== Console output =====*/
	if( nConsoleHeight > 0 ) {
		// Print clipping histogram(s)
		if( vClipHistoL.size() + vClipHistoR.size() > 0 ) {
			unsigned int nWidth  = nConsoleWidth,
			             nHeight = nConsoleHeight/2;
			if(bStereo)
				nWidth /= 2;
			vector<unsigned long> vnScaledL(vClipHistoL);
			vector<unsigned long> vnScaledR;
			if(bStereo) {
				vnScaledR = vClipHistoR;
				// ensure symmetry for stereo drawing
				if( vnScaledR.size() > vnScaledL.size() )
					vnScaledL.resize( vnScaledR.size(), 0 );
				else if( vnScaledL.size() > vnScaledR.size() )
					vnScaledR.resize( vnScaledL.size(), 0 );
			}
			vnScaledL = SoundStats::resampleHisto( vnScaledL, nWidth );
			if(bStereo)
				vnScaledR = SoundStats::resampleHisto( vnScaledR, nWidth );

			vector< vector<char> > vvcClipImage, vvcClipImageR;
			unsigned long nScale = vectorMax(vnScaledL);
			if(bStereo) {
				unsigned long nScaleR = vectorMax(vnScaledR);
				if( nScaleR > nScale )
					nScale = nScaleR;
			}
			if( nScale < nHeight )
				nScale = nHeight;
			// Render left, flip if stereo display
			vvcClipImage = renderHisto( vnScaledL, nScale, nHeight, bStereo );

			if(bStereo) {
				// Render right and glue to left
				vector< vector<char> > vvcClipImageR =
					renderHisto( vnScaledR, nScale, nHeight );
				for( unsigned int i=0; i<nHeight; i++ )
					vvcClipImage[i].insert( vvcClipImage[i].end(), vvcClipImageR[i].begin(), vvcClipImageR[i].end() );
			}
			squashASCIIImage(vvcClipImage);

			cout << "=== Clipping histogram (range: " << nScale << ") ===" << endl;
			for( unsigned int i=0; i<nHeight; i++ ) {
				for( unsigned int j=0; j<vvcClipImage[i].size(); j++ )
					cout << vvcClipImage[i][j];
				cout << endl;
			}
		}		
	
		// Print sample value histogram(s)
		vector<unsigned long> vHistoConsoleL = statsLeft.getHistogram( -1.0f, 1.0f, nConsoleHeight, true );
		vector<unsigned long> vHistoConsoleR;
		if( nGlobalChannels > 1 ) {
			vHistoConsoleR = statsRight.getHistogram( -1.0f, 1.0f, nConsoleHeight, true );
			if(bMonoOutput) {
				for( unsigned int i=0; i<vHistoConsoleL.size(); i++ )
					vHistoConsoleL[i] += vHistoConsoleR[i];
			}
		}

		// Print log histo on console output
		vector<float> vHistoLogL( nConsoleHeight, 0.0f ),
		              vHistoLogR( nConsoleHeight, 0.0f );
		double dLogHistoMax = 0;
		for( unsigned int i=0; i<nConsoleHeight; i++ ) {
			float fLog = log(1+vHistoConsoleL[i]);
			vHistoLogL[i] = fLog;
			if( fLog > dLogHistoMax )
				dLogHistoMax = fLog;
		}
		if(bStereo) {
			for( unsigned int i=0; i<nConsoleHeight; i++ ) {
				float fLog = log(1+vHistoConsoleR[i]);
				vHistoLogR[i] = fLog;
				if( fLog > dLogHistoMax )
					dLogHistoMax = fLog;
			}
		}

		cout << "=== Amplitude log histogram ===" << endl;
		if(!bStereo) {
			double dScale = (double)nConsoleWidth/dLogHistoMax;
			for( unsigned int i=0; i<nConsoleHeight; i++ ) {
				int nOut = (int)round(vHistoLogL[i]*dScale);
				for( int j=0; j<nOut; j++ )
					cout << "*";
				cout << endl;
			}
		}
		else { // Stereo display: mirrored histograms
			unsigned int nHalfConsoleWidth = nConsoleWidth/2;
			double dScale = (double)nHalfConsoleWidth/dLogHistoMax;
			for( unsigned int i=0; i<nConsoleHeight; i++ ) {
				int nOutL = (int)round(vHistoLogL[i]*dScale),
				    nOutR = (int)round(vHistoLogR[i]*dScale);
				for( int j=0; j<(int)(nHalfConsoleWidth-nOutL); j++ )
					cout << " ";
				for( int j=0; j<nOutL; j++ )
					cout << "*";
				for( int j=0; j<nOutR; j++ )
					cout << "*";
				cout << endl;
			}
		}
		cout << "===============================" << endl;
	}


	/*===== File output =====*/
	if(sOutFile) {
		string sHistoFile(sOutFile);
		if(bStereo)
			sHistoFile += "-samplesL.txt";
		else
			sHistoFile += "-samples.txt";
		writeHistoFile( sHistoFile, vRawHistoNegL, vRawHistoPosL );
		if(bStereo) {
			sHistoFile = string(sOutFile)+"-samplesR.txt";
			writeHistoFile( sHistoFile, vRawHistoNegR, vRawHistoPosR );
		}

		if( vClipHistoL.size() + vClipHistoR.size() > 0 ) {
			if(bStereo)
				sHistoFile = string(sOutFile) + "-clippingL.txt";
			else
				sHistoFile = string(sOutFile) + "-clipping.txt";
			ofstream oFile( sHistoFile.c_str() );
			for( unsigned long i=0; i<vClipHistoL.size(); i++ )
				oFile << i << '\t' << vClipHistoL[i] << "\n";
			oFile.close();
		
			if(bStereo) {
				sHistoFile = string(sOutFile) + "-clippingR.txt";
				oFile.open( sHistoFile.c_str() );
				for( unsigned long i=0; i<vClipHistoR.size(); i++ )
					oFile << i << '\t' << vClipHistoR[i] << "\n";
				oFile.close();
			}
		}
	}

	return 0;
}
