/*
 *  decoder.cpp
 *  Simple MP3 decoder class, frontend to libMAD
 *
 *  Created by Alexander Thomas on 2006/07/21.
 *  C++ version: 2011/01/04
 *  Last modified on 2011/01/15.
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2011, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
using namespace std;
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <cassert>

#include <id3tag.h>
#include "mp3decoder.h"

//#define DEEP_DEBUG

/* Length of input buffer  */
#define BUF_SIZE 32768
#define MYABS(x) (x < 0 ? -x : x)


MP3Decoder::MP3Decoder() :
	m_file(NULL),
	m_stream(NULL),
	m_frame(NULL),
	m_synth(NULL)
{
	m_nMaxDecode = 0;
	m_nSamplesRequested = 0;
	m_inBuffer.pData = NULL;
	m_outBuffer.bValid = false;
	m_outBuffer.pSamplesLeft  = NULL;
	m_outBuffer.pSamplesRight = NULL;
	m_outBuffer.nSamples = 0;
	m_nOutBufferSize = 0;
}

MP3Decoder::MP3Decoder( FILE *file, unsigned long fSize )
{
	init( file, fSize );

	m_nSamplesRequested = 0;
	m_outBuffer.bValid = false;
	m_outBuffer.pSamplesLeft  = NULL;
	m_outBuffer.pSamplesRight = NULL;
	m_outBuffer.nSamples = 0;
	m_nOutBufferSize = 0;
}



MP3Decoder::~MP3Decoder()
{
	finish();
}


void MP3Decoder::init( FILE *file, unsigned long fSize )
{
	m_file = file;
	m_nMaxDecode = fSize;
	m_nRead = 0;

	m_inBuffer.length = 0;
	m_inBuffer.pData = new unsigned char[BUF_SIZE];

	// Don't feel tempted to replace these with non-pointer member variables.
	m_stream = new mad_stream;
	m_frame  = new mad_frame;
	m_synth  = new mad_synth;
	
  mad_stream_init(m_stream);
  mad_frame_init(m_frame);
  mad_synth_init(m_synth);

	// Obviously, a freshly created stream will have nothing in its buffer
	m_stream->error = MAD_ERROR_BUFLEN;
}


void MP3Decoder::finish()
{
	if(m_stream) {
		mad_synth_finish(m_synth);
		mad_frame_finish(m_frame);
		mad_stream_finish(m_stream);
		m_stream = NULL;
		m_frame  = NULL;
		m_synth  = NULL;
	}
	if(m_inBuffer.pData)
		delete [] m_inBuffer.pData;
	m_inBuffer.pData = NULL;
	if(m_outBuffer.pSamplesLeft)
		delete [] m_outBuffer.pSamplesLeft;
	m_outBuffer.pSamplesLeft = NULL;
	if(m_outBuffer.pSamplesRight)
		delete [] m_outBuffer.pSamplesRight;
	m_outBuffer.pSamplesRight = NULL;
}


unsigned int MP3Decoder::getNumChannels()
{
	assert( m_inBuffer.pData );
	if( ! m_outBuffer.bValid )
		run();
	return m_outBuffer.nChannels;
}

long MP3Decoder::getSampleRate()
{
	assert( m_inBuffer.pData );
	if( ! m_outBuffer.bValid )
		run();
	return m_outBuffer.nSampleRate;
}


bool MP3Decoder::run( unsigned long nSamples )
{
	assert( m_inBuffer.pData );
#ifdef DEEP_DEBUG
	cout << "run called with nSamples=" << nSamples << ", samples in buffer: " << m_outBuffer.nSamples << endl;
#endif
	// Originally I had implemented this with mad_decoder_run. However, it cannot
	// be paused and resumed because it resets everything and purges its buffers
	// when stopping. Therefore I reimplemented the entire run function.

	m_nSamplesRequested = nSamples;
	bool bError = false, bExit = false;

	do {
		if( m_stream->error == MAD_ERROR_BUFLEN ) {
			if( !getInput() )
				break;
		}
		while(true) {
			if( mad_frame_decode(m_frame, m_stream) == -1 ) {
				if( m_stream->error == MAD_ERROR_BUFLEN )
					break;
				printError();
				if( MAD_RECOVERABLE(m_stream->error) ) { // mute or skip? Maybe provide option
					bError = true;
					mad_frame_mute(m_frame);
				}
				else {
					bError = true;
					bExit = true;
					break;
				}
			}
			else
				// Even though it doesn't make sense, I need to reset this manually.
				// Otherwise everything goes haywire.
				m_stream->error = MAD_ERROR_NONE;

			mad_synth_frame(m_synth, m_frame);
			sendOutput( &m_synth->pcm );

			if( m_outBuffer.nSamples >= m_nSamplesRequested ) {
				bExit = true;
				break;
			}
		}

	} while( !bExit && m_stream->error == MAD_ERROR_BUFLEN );

	m_nSamplesRequested = 0;

	return bError;
}


bool MP3Decoder::getSamples( outBuffer &output, unsigned long nSamples )
{
	assert( m_inBuffer.pData );
	assert( output.pSamplesLeft );
#ifdef DEEP_DEBUG
	cout << "getSamples called with nSamples=" << nSamples << ", samples in buffer: " << m_outBuffer.nSamples << endl;
#endif

	// Get more decoded samples if necessary
	if( m_outBuffer.nSamples < nSamples )
		run( nSamples-m_outBuffer.nSamples );

	assert( output.nChannels == m_outBuffer.nChannels );

	unsigned long nDataSize = nSamples;
	if( m_outBuffer.nSamples < nDataSize ) // should only happen at EOF
		nDataSize = m_outBuffer.nSamples;

	// As with inputCB, using memmove is inefficient, could use cyclic buffers to avoid it
	memcpy( output.pSamplesLeft, m_outBuffer.pSamplesLeft, nDataSize*sizeof(float) );
	unsigned long nRestSize = (m_outBuffer.nSamples-nDataSize)*sizeof(float);
	memmove( m_outBuffer.pSamplesLeft, &(m_outBuffer.pSamplesLeft[nDataSize]), nRestSize );
	if( m_outBuffer.nChannels > 1 ) {
		memcpy( output.pSamplesRight, m_outBuffer.pSamplesRight, nDataSize*sizeof(float) );
		memmove( m_outBuffer.pSamplesRight, &(m_outBuffer.pSamplesRight[nDataSize]), nRestSize );
	}
	m_outBuffer.nSamples -= nDataSize;

	output.bValid = true;
	output.nSampleRate = m_outBuffer.nSampleRate;
	output.nSamples = nDataSize;
	
	return (nDataSize > 0);
}


bool MP3Decoder::getID3Info( const char* filePath,
                             long &id3_length, bool &hasV1, bool &hasV2 )
{
	unsigned char firstBuf[1024], lastBuf[128];
		
	/* Using id3_file_open and id3_file_tag appears to be useless, because it
	 * gives wrong information about the tag version and size. Or, I just have
	 * no idea how to use it correctly due to the total absence of documentation.
	 * So I do it the low-level way, which gives a lot more control too. */
	unsigned long bufLength = readFirstPart( firstBuf, 1024, filePath );
	if( bufLength == 0 )
		return false;
		
	readLastPart( lastBuf, 128, filePath );
	if( id3_tag_query( lastBuf, 128 ) == 128 )
		hasV1 = true;
	/* TODO: allow ID3v2 footers: use negative numbers */
	id3_length = id3_tag_query( firstBuf, bufLength );
	if( id3_length > 0 )
		hasV2 = true;

	return true;
}


inline signed int MP3Decoder::scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));
	
  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;
	
  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}


void MP3Decoder::printStatus() const
{
	cout << "Read input bytes: " << m_nRead << endl
	     << "  Current output buffer allocation: " << m_nOutBufferSize << " samples" << endl
	     << "  Samples in output buffer: " << m_outBuffer.nSamples << endl
	     << "  Bytes in input buffer: " << m_inBuffer.length << endl
	     << "    of which undecoded: " << (m_stream->bufend - m_stream->next_frame) << endl;
}


bool MP3Decoder::getInput()
{
	/* Move remaining part of buffer to the front.
	 * I use memmove, and copying large chunks of memory is generally a bad idea,
	 * but in this case the chunk should always be rather small. */
	unsigned long restLength = m_stream->bufend - m_stream->next_frame;
#ifdef DEEP_DEBUG
	fprintf( stderr, "getInput: %lu bytes left over from last decode\n", restLength );
#endif
	if( restLength > 0 )
		memmove( m_inBuffer.pData, m_inBuffer.pData+m_inBuffer.length-restLength, restLength );
	
	/* Fill rest of buffer with new data */
	unsigned long toRead = BUF_SIZE-restLength;
	if(m_nMaxDecode) {
		if( m_nRead >= m_nMaxDecode )
			return false;
		else if( m_nRead+toRead > m_nMaxDecode )
			toRead = m_nMaxDecode-m_nRead;
	}
	
	unsigned long newLength = fread( m_inBuffer.pData+restLength, 1, toRead, m_file );
	m_nRead += newLength;
	if( newLength <= 0 )
    return false;
	m_inBuffer.length = restLength+newLength;
	
#ifdef DEEP_DEBUG
	fprintf( stderr, "inputCB: Read %lu bytes, buffer: %lu\n", newLength, m_inBuffer.length );
#endif
	mad_stream_buffer( m_stream, m_inBuffer.pData, m_inBuffer.length );
	
	return true;
}


void MP3Decoder::sendOutput( struct mad_pcm *pcm )
{
  unsigned int nChannels = pcm->channels;
  unsigned int nSamples  = pcm->length;
  const mad_fixed_t *left_ch  = pcm->samples[0];
 	const mad_fixed_t *right_ch = pcm->samples[1];

	if( ! m_outBuffer.bValid ) {
		m_outBuffer.bValid      = true;
		m_outBuffer.nChannels   = nChannels;
		m_outBuffer.nSampleRate = pcm->samplerate;
	}
	
#ifdef DEEP_DEBUG
	fprintf( stderr, "sendOutput: Decoded %u samples\n", nSamples );
#endif
	
	/* Extend output buffer if necessary */
	if( m_nOutBufferSize < m_outBuffer.nSamples+nSamples ) {
		unsigned long nNewSize = m_outBuffer.nSamples+nSamples+8192; // some extra margin, TODO: tune
		float *newBuffLeft = new float[nNewSize];
		if(m_outBuffer.pSamplesLeft) {
			memcpy( newBuffLeft, m_outBuffer.pSamplesLeft, m_outBuffer.nSamples*sizeof(float) );
			delete [] m_outBuffer.pSamplesLeft;
		}
		float *newBuffRight=NULL;
		if( nChannels > 1 ) {
			newBuffRight = new float[nNewSize];
			if(m_outBuffer.pSamplesRight) {
				memcpy( newBuffRight, m_outBuffer.pSamplesRight, m_outBuffer.nSamples*sizeof(float) );
				delete [] m_outBuffer.pSamplesRight;
			}
		}

		m_outBuffer.pSamplesLeft  = newBuffLeft;
		m_outBuffer.pSamplesRight = newBuffRight;
		m_nOutBufferSize = nNewSize;
	}
	unsigned long nStart = m_outBuffer.nSamples;
	m_outBuffer.nSamples += nSamples;
	/* Append samples in the range -1,1 (with potential overshoots) to output buffer. */
  for( unsigned long i=nStart; i<m_outBuffer.nSamples; i++ ) {
		m_outBuffer.pSamplesLeft[i] = (float)(*left_ch++) / (float)(1L << MAD_F_FRACBITS);
    if( nChannels > 1 )
			m_outBuffer.pSamplesRight[i] = (float)(*right_ch++) / (float)(1L << MAD_F_FRACBITS);
  }
}


void MP3Decoder::printError()
{
	// Not sure if the byte offset is correct as calculated here, checkMe
  fprintf( stderr, "decoding error 0x%04x (%s) at byte offset 0x%lx\n",
	         m_stream->error, mad_stream_errorstr(m_stream),
	         m_nRead - (long)(m_stream->this_frame - m_inBuffer.pData) );
}


int MP3Decoder::readFirstPart( unsigned char *buff, unsigned int length,
                               const char *fileName )
{
	int len;
	FILE *file = fopen( fileName, "rb" );
	if( file == NULL )
		return -1;
	len = fread( buff, 1, length, file );
	fclose(file);
	
	return len;
}

int MP3Decoder::readLastPart( unsigned char *buff, unsigned int length,
                              const char *fileName )
{
	int len;
	FILE *file = fopen( fileName, "rb" );
	if( file == NULL )
		return -1;
	if( fseek( file, -length, SEEK_END ) ) {
		fclose(file);
		return -1;
	}
	len = fread( buff, 1, length, file );
	fclose(file);
	
	return len;
}
