/*
 *  soundfile.cpp
 *  An abstract class for reading or writing audio files.
 *
 *  Created by Alexander Thomas on 2008/01/22.
 *  Modified on: 2014/01/11
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2014, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <cmath>
// for uint32_t etc. Ideally this should be <cstdint>
#include <stdint.h>
using namespace std;
#include "endianconv.h"
#include "soundfile.h"


SoundFile::SoundFile( const string &sFileName, bool bVerbose ) :
	m_bVerbose (bVerbose),
	m_sFileName(sFileName)
{
	m_soundInfo.numChannels = 0;
	m_soundInfo.numSampleFrames = 0;
	m_soundInfo.sampleSize = 0;
	m_soundInfo.sampleRate = 0;
	m_soundInfo.totalDataSize = 0;
}


/* This is the most exact way to do it and there is no point in using faster
 * hacks with more inaccurate results because this will only be called a
 * few times in normal applications. */
double SoundFile::IEEE754ToDouble( extended_t ieee745 ) const
{
	char *buffer = (char*)ieee745;
	uint32_t mant1 = *((uint32_t *)(buffer+2));
	uint32_t mant2 = *((uint32_t *)(buffer+6));
	uint16_t exp   = *((uint16_t *)buffer);
	
	getFromBigEndian( &mant1, sizeof(uint32_t) );
	getFromBigEndian( &mant2, sizeof(uint32_t) );
	getFromBigEndian( &exp,   sizeof(uint16_t) );
	
	double result  = (double)mant1*pow(2.0,exp-16414);
	result += (double)mant2*pow(2.0,exp-16446);
	return result;
}


/* Same comment as with IEEE754ToDouble. */
void SoundFile::doubleToIEEE754( extended_t ieee754, double value ) const
{
	uint16_t exp = (uint16_t)(log(value)/M_LN2)+16383;
	double temp = value/pow(2.0,exp-16414);
	uint32_t mant1 = (uint32_t)temp;
	temp = (temp-mant1)*4294967296.0;
	uint32_t mant2 = (uint32_t)temp;
	
	getFromBigEndian( &mant1, sizeof(uint32_t) );
	getFromBigEndian( &mant2, sizeof(uint32_t) );
	getFromBigEndian( &exp,   sizeof(uint16_t) );
	
	uint16_t *buffer = (uint16_t*)ieee754;
	buffer[0] = exp;
	uint32_t *buff2 = (uint32_t *)&(buffer[1]);
	buff2[0] = mant1;
	buff2[1] = mant2;	
}


bool SoundFile::scanFile( istream &inStream, const char *pSearch,
                          unsigned int nSize )
{
	int *pBuff = new int[nSize]; // cyclic buffer
	unsigned int nCycloPos = 0;
	bool bNotFound;
	
	do {
		pBuff[nCycloPos] = inStream.get();
		if( pBuff[nCycloPos] == EOF ) {
			delete [] pBuff;
			return false;
		}
		nCycloPos = (nCycloPos+1) % nSize;
		
		bNotFound = false;
		for( unsigned int i=0; i<nSize; i++ ) {
			if( pBuff[(nCycloPos+i)%nSize] != pSearch[i] ) {
				bNotFound = true;
				break;
			}
		}
	}
	while( bNotFound );
	
	delete [] pBuff;
	
	return true;
}	
