/*
 *  soundfile.h
 *  An abstract class for reading or writing audio files.
 *
 *  Created by Alexander Thomas on 2008/01/22.
 *  Modified on 2011/01/18
 * Released under the New BSD License
 * Contact: doctor.lex at gmail.com
 *
 * Copyright (c) 2011, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SOUNDFILE_H
#define SOUNDFILE_H

#include <iostream>
#include <fstream>
#include <string>
using namespace std;

/* 80 bit IEEE Standard 754 floating point number */
typedef unsigned char extended_t[10];

enum SoundFileMode_t { Read, Write };
enum SoundFileState_t { Good, Bad, Eof };

typedef struct {
	short  numChannels;
	unsigned long numSampleFrames;
	short  sampleSize;
	double sampleRate;
	unsigned long totalDataSize;
} soundFileInfo_t;


class SoundFile
{
public:
	/* Create a new SoundFile, connected to the file sFileName on disk.
	 * If bVerbose, error messages will be printed to stderr. */
	SoundFile( const string &sFileName, bool bVerbose=false );
	
	/* Returns SoundFileMode_t depending on the subclass. */
	virtual SoundFileMode_t fileMode() const = 0;
	
	/* Return the file on disk associated with this SoundFile. */
	string getFileName() const { return m_sFileName; }
	
	/* Returns the state of the file:
	 * Good means the file is open and ready for reading/writing
	 * Bad means the file does not exist, cannot be opened, or is not supported
	 * Eof means the file is OK, but the read pointer is at the end of the sound chunk
	 * or the Write file has been closed for writing. */
	SoundFileState_t getFileState() const { return m_fileState; }
	
	/* Get the number of sample frames in this file. */
	unsigned long getNumSampleFrames() const { return m_soundInfo.numSampleFrames; }
	
	/* Get the number of channels in this file. */
	short getNumChannels() const { return m_soundInfo.numChannels; }

	/* Get the sample rate of this file. */
	double getSampleRate() const { return m_soundInfo.sampleRate; }
	
	/* Get the sample size of this file (in bits). */
	short getSampleSize() const { return m_soundInfo.sampleSize; }
	
	/* Get the number of sample frames in this file. */
	unsigned long getNumFrames() const { return m_soundInfo.numSampleFrames; }

	/* Get the duration in number of seconds of this file. */
	double getDuration() const { return getNumFrames()/getSampleRate(); }

	/* Get the position (frame number) of the read/write pointer in the file.
	 * This function cannot be declared const because tellg() and tellp() are
	 * not const for reasons unknown. */
	virtual unsigned long getFilePosition() = 0;

	void setVerbose( bool bVerbose=true ) { m_bVerbose = bVerbose; }
	bool getVerbose() const { return m_bVerbose; }

protected:
	/* Converts an 80 bit IEEE Standard 754 floating point number to a double. */
	double IEEE754ToDouble( extended_t buffer ) const;
	
	/* Converts a double to 80 bit IEEE Standard 754 floating point number. */
	void doubleToIEEE754( extended_t ieee754, double value ) const;
	
	/* Scan file for a search pattern of nSize bytes.
	 * Returns true if found, false if not.
	 * If found, the stream pointer will be right after the search pattern.*/
	static bool scanFile( istream &inStream, const char *pSearch,
	                      unsigned int nSize );
	
	bool   m_bVerbose;
	string m_sFileName;
	soundFileInfo_t  m_soundInfo;
	SoundFileState_t m_fileState;
};

#endif //SOUNDFILE_H
