/*
 * soundstats.h
 * A class for collecting statistics about an audio stream
 * represented by floating-point values between -1 and 1.
 * A SoundStats object records a histogram of the sample values
 *   quantized to 16-bit, as well as a histogram of how many
 *   consecutive samples clipped.
 *
 * Created by Alexander Thomas on 2011/01/15
 * Last modification: 2011/06/03
 * Released under the New BSD License
 *
 * Copyright (c) 2011, Alexander Thomas
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Alexander Thomas nor the
 *       names of contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SOUNDSTATS_H
#define SOUNDSTATS_H

#include <vector>
using namespace std;

class SoundStats
{
public:
	/* Constructor */
	SoundStats();
	/* Copy constuctor */
	SoundStats( const SoundStats &stats );

	/* Set the threshold above which a sample is considered to be clipping, i.e.
	 *   a sample clips if it >= fThresh or sample <= -fThresh.
	 * fThresh must be strictly positive. */
	void setClipThreshold( float fThresh=0.99f );

	/* Return the clipping threshold. */
	float getClipThreshold() const { return m_fClipThreshold; }

	/* Collect statistics for a chunk of floating-point samples. */
	void collectStats( float *pfSamples, unsigned long nSamples );
	/* Collect statistics for a chunk of signed short (16-bit) samples. */
	void collectStats( short *pnSamples, unsigned long nSamples );

	/* This will finish any pending statistics and reset internal parameters
	 *   such that the next analyzed block of samples will not be assumed to
	 *   follow directly on the previous. */
	void finish();

	/* Clear all statistics. */
	void clear();

	/* Return the number of samples that have been processed. */
	unsigned long long getNumSamples() const { return m_nSamples; }

	/* Return the minimum and maximum sample values that were encountered */
	float getMinValue() const { return m_fMinValue; }
	float getMaxValue() const { return m_fMaxValue; }

	/* Return the sample value histograms. The histogram contains an entry for
	 *   each 16-bit quantized value, i.e. each bin is 2.0/65536 wide in the
	 *   [-1,1] range. The histogram is split up in positive values from 0 on
	 *   and negative values from -1 on. The histograms may contain more than
	 *   32768 values each, if the signal had amplitudes outside [-1,1]. */
	void getHistogram( vector<unsigned long> &vnHistogramPos,
	                   vector<unsigned long> &vnHistogramNeg ) const;

	/* Return a resampled histogram, with the given minimum and maximum values
	 *   and number of bins. The only restriction is that fMin must be negative
	 *   and fMax positive. If bClipValues is true, values outside the range
	 *   will be added to the first and last bins, otherwise they are ignored. */
	vector<unsigned long> getHistogram( float fMin=-1.0f, float fMax=1.0f,
	                                    unsigned int nBins=20,
	                                    bool bClipValues=true ) const;

	/* Return the clipping histogram.
	 * The element at position i in the vector contains the number of times
	 *   a series of clipping samples of length i+1 was encountered. */
	vector<unsigned long> getClipHistogram() const { return m_vnClipHistogram; }

	/* Return the same-sign clipping histogram.
	 * This is the same kind of histogram as above, but it only considers
	 *   series of clipping samples that have the same sign. */
	vector<unsigned long> getClipHistogram2() const { return m_vnClipHistogram2; }


	/* Resamples a histogram to nBins bins by averaging merged bins, if nBins is
	 * smaller than the size of the histogram. If nBins is larger, the histogram
	 * is returned unmodified but padded with zeros to make it nBins wide.
	 * If due to averaging a bin with at least one sample in it would be reduced to
	 * zero, it is forced to 1.
	 * (This is intended to be used for the clipping histogram) */
	static vector<unsigned long> resampleHisto( const vector<unsigned long> vnHisto,
	                                            unsigned int nBins );

private:
	// Update the clipping histograms for a new floating-point sample
	void updateClipping( float fValue );

	// This vector covers all positive samples
	vector<unsigned long> m_vnHistogramPos;
	// This vector covers all strictly negative samples
	vector<unsigned long> m_vnHistogramNeg;
	unsigned long long m_nSamples;

	float m_fClipThreshold;
	// Absolute clipping histogram
	vector<unsigned long> m_vnClipHistogram;
	// Same-sign clipping histogram
	vector<unsigned long> m_vnClipHistogram2;
	unsigned long m_nClipLength;
	unsigned long m_nClipLength2;
	float m_fLastClipValue;

	float m_fMinValue;
	float m_fMaxValue;
};


#endif // SOUNDSTATS_H
